/******************************************************************* 
* Copyright (c) 2010 by Hesine Technologies, Inc. 
* All rights reserved. 
* 
* This file is proprietary and confidential to Hesine Technologies. 
* No part of this file may be reproduced, stored, transmitted, 
* disclosed or used in any form or by any means other than as 
* expressly provided by the written permission from Jianhui Tao 
* 
* ****************************************************************/

#ifdef __MMI_NMS_MESSAGE__
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/uio.h>
#include <sys/time.h>
#include <semaphore.h>
#include <pthread.h>
#include <errno.h>
#include <netinet/in.h>
#include <pthread.h>
#include <netdb.h>
#include <sys/select.h>
#include <netinet/tcp.h>
#include <sys/poll.h>

#include<jni.h>

#include "hpnsPlatform.h"
#include "tmodule.h"
#include "hpnsAppEngine.h"
#include "hpnsNetwork.h"

/* auto increase socket id */
static UINT32  socketIncId = 1 ;

static int  savePreviousUdpFd = -1 ;
static int  savePreviousTcpFd = -1 ;

static UINT32 getSocketIncId() 
{
    /* because we konw the hpnsOpenUdpSocket and hpnsOpenTcpSocket always be called in the same ENGINE thread
     *   So it is not needed to add a MUTEX to protect it 
     */ 
    socketIncId++ ;
    return socketIncId ;
}


void hpnsSocketReset(SNmsSocket* nmsSocket) 
{
    nmsSocket->sockFd = -1 ;
    nmsSocket->incId    = (UINT32)-1 ;
}

int hpnsSocketIsValid(const SNmsSocket* nmsSocket) 
{
    if (nmsSocket->sockFd < 0 || nmsSocket->incId == (UINT32)-1 || nmsSocket->incId == 0)
        return 0 ;

    return 1 ;
}

int hpnsSocketIsEqual(const SNmsSocket* socket1, const SNmsSocket* socket2) 
{
    if (!hpnsSocketIsValid(socket1))
        return 0 ;

    return (socket1->sockFd == socket2->sockFd) && (socket1->incId == socket2->incId) ;
}

INT32 hpnsOpenUdpSocket(char *asyncFlag, SNmsSocket* nmsSocket)
{
	struct sockaddr_in localAddr;
	int                sockFd;
	struct hostent     *hp;
	char               name[100];
	int                len = 100;
	int                reuse, nocheck;
	int				   ttl = 128;

	if (( sockFd = socket (AF_INET, SOCK_DGRAM, 0)) < 0) {
		nprintf("failed to open udp socket, error: %s", strerror(errno));
		return -1;
	}

	reuse = 1;
	if ( setsockopt(sockFd, SOL_SOCKET, SO_REUSEADDR, (void *)&reuse, sizeof(reuse)) < 0 )  
	    nprintf("open udp setsockopt SO_REUSEADDR failed, error: %s", strerror(errno));

	ttl = 128;
	if ( setsockopt(sockFd, IPPROTO_IP, IP_TTL, &ttl, sizeof(ttl)) < 0) 
            nprintf("open udp setsockopt IP_TTL failed, error: %s", strerror(errno));
       
      nmsSocket->sockFd = sockFd ;
      nmsSocket->incId    = getSocketIncId() ;

	taosSendMsgToModule( &(moduleObj[1]), 0, 0, sizeof(SNmsSocket), (char*)nmsSocket);

	*asyncFlag = 0;

        nprintf("open udp socket: %d, incId: %d", nmsSocket->sockFd, nmsSocket->incId) ;
	
	return 0;
}

void hpnsCloseUdpSocket(SNmsSocket* nmsSocket)
{
    int        ret = -1; 

    nprintf("close udp socket: %d, incId: %d", nmsSocket->sockFd, nmsSocket->incId) ;
    
    ret = shutdown(nmsSocket->sockFd, 2);

    if (ret != 0)
        nprintf("shutdown udp sockId: %d, incId: %d socket return: %d, errno: %s", nmsSocket->sockFd, nmsSocket->incId, ret, strerror(errno));
    
    if (savePreviousUdpFd != -1) {
        ret = close(savePreviousUdpFd) ;
	 if (ret != 0)
            nprintf("close previous udp socket fd: %d, error return: %d, %s", savePreviousUdpFd, ret, strerror(ret)); 
    }

    savePreviousUdpFd = nmsSocket->sockFd ;  
}

int hpnsSendUdpData(SNmsSocket* nmsSocket, UINT32 ip, UINT16 port, UINT8 *msg, int msgLen) 
{
       UINT32 sockFd = nmsSocket->sockFd ;
	struct sockaddr_in destAdd;
	int                ret;

	destAdd.sin_family      = AF_INET;
	destAdd.sin_addr.s_addr = ip; 
	destAdd.sin_port        = port;

	ret = sendto( sockFd, msg, msgLen, MSG_DONTWAIT, (struct sockaddr *)&destAdd, sizeof(destAdd));
#ifdef __ANDROIDOS__
	if(ret < 0){
		nprintf("send udp data error,reason:%s, ret:%d",strerror(errno), ret);
	}
#endif
	
	return ret;
}

int hpnsRecvUdpData(SNmsSocket* nmsSocket, UINT8 *buffer, int bufferLen, UINT32 *ip, UINT16 *port)
{
       UINT32 sockFd = nmsSocket->sockFd ;
	int                dataLen = -1;
	struct sockaddr_in sourceAdd;
	int                sourceAddLen;	

	sourceAddLen = sizeof(sourceAdd);
	
	dataLen = recvfrom( sockFd, buffer, bufferLen, MSG_DONTWAIT, (struct sockaddr *) &sourceAdd, &sourceAddLen );
       
	if ( dataLen < 0 && errno == EINTR ) {
            nprintf("recv udp socket was intercept");
            taosSendMsgToModule( &(moduleObj[1]), 0, 0, sizeof(SNmsSocket), (char*)nmsSocket);
            return 0;
       }

        if ( dataLen < 0 ) {
            nprintf("recv udp socket error, reason:%s", strerror(errno));
            return dataLen ;
        }
         
	*ip = sourceAdd.sin_addr.s_addr;
	*port = sourceAdd.sin_port;
        taosSendMsgToModule( &(moduleObj[1]), 0, 0, sizeof(SNmsSocket), (char*)nmsSocket);
	
	return dataLen;
}

int  hpnsGetTtl()
{
	return 128;
}

int hpnsOpenTcpSocket(SNmsSocket* nmsSocket)
{
    int  sockFd  = -1 ;
    struct timeval tm = {0};
    int  tmpData = 0 ;
    
    if ((sockFd = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0) {
        nprintf("failed to open tcp socket, error: %s", strerror(errno));
        return -1 ;
    }
    
    tm.tv_sec  = 3; tm.tv_usec = 0;
    if (setsockopt(sockFd, (int)SOL_SOCKET, (int)SO_SNDTIMEO, (void*)&tm, sizeof(struct timeval)) < 0) 
        nprintf("failed to set SO_SNDTIMEO of tcp socket, error: %s", strerror(errno));

    tmpData = 1 ;
    if (setsockopt(sockFd, IPPROTO_TCP, TCP_QUICKACK, (void*)&tmpData, sizeof(int)) < 0)
        nprintf("failed to set TCP_QUICKACK of tcp socket, error: %s", strerror(errno));

    tmpData = 1 ;
    if (setsockopt(sockFd, IPPROTO_TCP, TCP_NODELAY, (void*)&tmpData, sizeof(int)) < 0)
        nprintf("failed to set TCP_NODELAY of tcp socket, error: %s", strerror(errno));

    /*  don't do that to kill the tcp TIME-WAIT status, it may got some unexpect result
    {
        struct linger lingerOption;
        lingerOption.l_onoff = 1;
        lingerOption.l_linger = 0;
        setsockopt(sockFd, SOL_SOCKET, SO_LINGER, (void*)&lingerOption, sizeof(lingerOption));
    } */
    
    /* can't get the local ip and port until the tcp connection is established
    {
        struct sockaddr_in addr = {0};
        socklen_t               nameLen = sizeof(addr);
        if (getsockname(sockFd, (struct sockaddr*)&addr, &nameLen) < 0) 
            nprintf("tcp socket getsockname errno %d, %s", errno, strerror(errno));
        else
            nprintf("tcp socket client  ip: %s, port:%d", inet_ntoa(addr.sin_addr), ntohs(addr.sin_port));
    } */
    
    nmsSocket->sockFd = sockFd; 
    nmsSocket->incId    = getSocketIncId() ;

    nprintf("open tcp socket: %d, incId: %d", nmsSocket->sockFd, nmsSocket->incId) ;
    
    return 0;
}

void  hpnsCloseTcpSocket(SNmsSocket* nmsSocket)
{
    int ret = 0 ;

    nprintf("close tcp socket: %d, incId: %d", nmsSocket->sockFd, nmsSocket->incId) ;
       
    ret = shutdown(nmsSocket->sockFd, 2);
  
    if (ret != 0) 
        nprintf("shutdown tcp socket error return: %d, %s", ret, strerror(ret));

    if (savePreviousTcpFd != -1) {
        ret = close(savePreviousTcpFd) ;
	 if (ret != 0)
            nprintf("close previous tcp socket fd: %d, error return: %d, %s", savePreviousTcpFd, ret, strerror(ret)); 
    }

    savePreviousTcpFd = nmsSocket->sockFd ;
}

static int connectTcpSocket(SNmsSocket* nmsSocket, UINT32 destIp, UINT16 destPort, UINT32 timeOut)
{
    struct sockaddr_in serverAddr;
    int ret = -1 ;
    struct timeval tm = {0} ;

    struct pollfd really;
    int junk;
    socklen_t junk_more;
    int pollTimeOut = 0 ;
        
    memset((char *) &serverAddr, 0, sizeof (serverAddr));
    serverAddr.sin_family = AF_INET;
    serverAddr.sin_addr.s_addr = destIp;
    serverAddr.sin_port = destPort;

    tm.tv_sec  = timeOut; tm.tv_usec = 0;
    setsockopt(nmsSocket->sockFd, (int)SOL_SOCKET, (int)SO_SNDTIMEO, (void*)&tm, sizeof(struct timeval));


//modify for pn-169
#if 1
    ret = connect(nmsSocket->sockFd, (struct sockaddr *)&serverAddr, sizeof(serverAddr));
    if( ret >= 0) {
        nprintf("tcp connect is ok");
        ret = 1 ;
        goto _exit ;
    }
    if(ret == -1)
    {
    	if(errno != EINTR)
    	{
	        nprintf("tcp connect -- 0 EINTR");
    		goto _error;
    	}

    	really.fd = nmsSocket->sockFd;
    	really.events = POLLOUT;

		pollTimeOut = timeOut*1000;
    	while ( poll (&really, 1, pollTimeOut) == -1 )
    		if ( errno != EINTR )
    		{
    	        nprintf("tcp connect -- 1 poll");
    			goto _error;
    		}

    	junk_more = sizeof(junk);
        if ( getsockopt (nmsSocket->sockFd, SOL_SOCKET, SO_ERROR,
                         &junk,
                         &junk_more) == -1 )
          {
	        nprintf("tcp connect -- 2 getsockopt");
			goto _error;
          }

        if ( junk != 0 )
          {
	        nprintf("tcp connect -- 3 junk: %s",strerror (junk_more));
			goto _error;
          }
    }

    ret  = 1;
    goto _exit ;

#else
    while (connect(nmsSocket->sockFd, (struct sockaddr *)&serverAddr, sizeof(serverAddr)) == -1 && errno != EISCONN )
    if ( errno == EINTR )
    {
        nprintf("tcp connect is ok");
        ret = 1 ;
        goto _exit ;
    }
#endif
//modify end!
_error:
	ret = -1;
    nprintf("tcp connect is failed, errno %d, %s , ret:%d", errno, strerror(errno),ret);

_exit:

    tm.tv_sec  = 3; tm.tv_usec = 0;
    setsockopt(nmsSocket->sockFd, (int)SOL_SOCKET, (int)SO_SNDTIMEO, (void*)&tm, sizeof(struct timeval));

    {
        struct sockaddr_in addr = {0};
        socklen_t nameLen = sizeof(addr);
        if (getsockname(nmsSocket->sockFd, (struct sockaddr*)&addr, &nameLen) < 0) 
            nprintf("tcp socket getsockname errno %d, %s", errno, strerror(errno));
        else
            nprintf("tcp socket get client ip:%s, port:%d", inet_ntoa(addr.sin_addr), ntohs(addr.sin_port));
    }
    
    if (ret > 0) 
        taosSendMsgToModule( &(moduleObj[3]), 0, 0, sizeof(SNmsSocket), (char*)nmsSocket);

    return ret ;
}



#define NMS_TCP_CONNECT_SYNC_TIMEOUT  5
#define NMS_TCP_CONNECT_ASYNC_TIMEOUT 30    /* TODO: sometimes got the 115(Operation now in progress) error even I set the tcp connect time out time to 120 seconds */

// #define NMS_TCP_CONNECT_SYNC 

#ifdef NMS_TCP_CONNECT_SYNC 

int hpnsConnectTcpSocket(SNmsSocket* nmsSocket, UINT32 destIp, UINT16 destPort)
{
    return connectTcpSocket(nmsSocket, destIp, destPort, NMS_TCP_CONNECT_SYNC_TIMEOUT) ;
}

#else

typedef struct {
    SNmsSocket  nmsSocket ;
    UINT32 ip;
    UINT16 port;
}TcpConnectThreadData ;

extern JavaVM* g_JavaVm ;
extern jclass g_classAdp ;
    
static void* tcpConnectThread(void *param) 
{
    TcpConnectThreadData* data = (TcpConnectThreadData*)param ;
    int ret = -1 ;
    JNIEnv* Env = NULL;

    jmethodID fget = NULL ;
    jmethodID frelease = NULL ;
        
    if (!data) {
        nprintf("error to connect tcp in thread for data is null") ;
        return NULL ;
    }
    
    if (!g_JavaVm) {
        nprintf("error to connect tcp in thread for g_JavaVm is null") ;
        return NULL ;
    }
    
    (*g_JavaVm)->AttachCurrentThread(g_JavaVm, &Env, NULL);

    nprintf("tcp module try to connect in thread, sockFd: %d, incId: %d", data->nmsSocket.sockFd, data->nmsSocket.incId) ;

    fget = (*Env)->GetStaticMethodID(Env, g_classAdp, "nmsGetWeakupLock","(I)V");
    frelease = (*Env)->GetStaticMethodID(Env, g_classAdp, "nmsReleaseWeakupLock","(I)V");

    if (fget) 
        (*Env)->CallStaticVoidMethod(Env,g_classAdp,fget, 0);
    else
        nprintf("error to get the fget function int tcp connect thread") ;

    ret = connectTcpSocket(&data->nmsSocket, data->ip, data->port, NMS_TCP_CONNECT_ASYNC_TIMEOUT) ;
    
    hpnsSendMsgToEngine((ret > 0) ? HPNS_MSG_DATA_WRITE_IND : HPNS_MSG_NW_ERROR_IND, (UINT8*)&(data->nmsSocket), sizeof(SNmsSocket));
    
    nprintf("tcp module connect in thread, ret: %d, sockFd: %d, incId: %d", ret, data->nmsSocket.sockFd, data->nmsSocket.incId) ;
    
    hpnsFreeL(data) ;

    if (frelease)
        (*Env)->CallStaticVoidMethod(Env, g_classAdp,frelease, 0);
    else
        nprintf("error to get the frelease function int tcp connect thread") ;
 
    (*g_JavaVm)->DetachCurrentThread(g_JavaVm);
           
    return NULL ;
}
    

int hpnsConnectTcpSocket(SNmsSocket* nmsSocket, UINT32 destIp, UINT16 destPort)
{
    TcpConnectThreadData* data = NULL ;
    pthread_t thread;    
    pthread_attr_t attr;

    data = (TcpConnectThreadData*)hpnsMallocL(sizeof(TcpConnectThreadData)) ;
    memset(data, 0, sizeof(TcpConnectThreadData)) ;

    data->nmsSocket = *nmsSocket ;
    data->ip              = destIp ;
    data->port           = destPort ;
    
    pthread_attr_init (&attr);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
    
    if (pthread_create(&thread, &attr, tcpConnectThread, (void*)data) != 0) {
        nprintf("tcp failed to create connect thread, but we still try to connect it by sync, sockFd: %d, incId: %d", nmsSocket->sockFd, nmsSocket->incId);
        return connectTcpSocket(nmsSocket, destIp, destPort, NMS_TCP_CONNECT_SYNC_TIMEOUT);
    }
    
    return 0; /* 0 mean async */
}

#endif  /* #ifdef NMS_TCP_CONNECT_SYNC  */


int hpnsSendTcpData(SNmsSocket* nmsSocket, UINT8 *msg, int msgLen)
{
    UINT32 sockFd = nmsSocket->sockFd ;
    int ret = -1;
    
    ret = send(sockFd, msg, msgLen, 0);

    if(ret < msgLen)
	nprintf("tcp socket error when sending, ret: %d, msgLen: %d, reason:%s", ret, msgLen, strerror(errno));
	
    return ret;
}

int hpnsRecvTcpData(SNmsSocket* nmsSocket, UINT8 *msg, int msgLen)
{
    UINT32 sockFd = nmsSocket->sockFd ;
    int ret = -1;
    

    ret = recv(sockFd, msg, msgLen, MSG_DONTWAIT);
    if(ret < 0)
        nprintf("tcp socket error when receiving, reason:%s", strerror(errno));

    return ret;
}


#endif

